#include "Plane3.h"
#include "ViewCellBsp.h"
#include "Mesh.h"

#include "ViewCell.h"
#include <stack>

//namespace GtpVisibilityPreprocessor {
/****************************************************************/
/*                  class BspNode implementation                */
/****************************************************************/
bool BspNode::IsRoot() const 
{
	return mParent == NULL;
}

/****************************************************************/
/*              class BspInterior implementation                */
/****************************************************************/

bool BspInterior::IsLeaf() const
{ 
	return false; 
}

void BspInterior::ReplaceChildLink(BspNode *oldChild, BspNode *newChild) 
{
	if (mBack == oldChild)
	{
		mBack = newChild;
	}
    else
	{
		mFront = newChild;
	}
}

void BspInterior::SetupChildLinks(BspNode *b, BspNode *f) 
{
    mBack = b;
    mFront = f;
}

/****************************************************************/
/*                  class BspLeaf implementation                */
/****************************************************************/

BspLeaf::BspLeaf(ViewCell *viewCell): mViewCell(viewCell) 
{
}

bool BspLeaf::IsLeaf() const 
{ 
	return true; 
}

/****************************************************************/
/*                  class BspTree implementation                */
/****************************************************************/

BspTree::BspTree(ViewCell *cell) 
{
	mRootCell = cell;
	mRoot = new BspLeaf(mRootCell);
}

  
void BspTree::Subdivide()
{
	BspNode *currentNode = mRoot;
  
	std::stack<BspTraversalData> tStack;
	//  stack<STraversalData> tStack;

	//tStack.push(tdata);

	while (!tStack.empty()) 
	{
	    BspTraversalData data = tStack.top();
	    tStack.pop();
    
		Mesh backPolys;
		Mesh frontPolys;

		BspNode *node = SubdivideNode(dynamic_cast<BspLeaf *>(data.mNode),
									  data.mParent,
									  &data.mViewCell,
								      data.mDepth,
									  backPolys,
									  frontPolys);
	
		if (!node->IsLeaf())
		{
			BspInterior *interior = dynamic_cast<BspInterior *>(node);

			// push the children on the stack
			if (interior->GetBack())
				tStack.push(BspTraversalData(interior->GetBack(), interior, backPolys, data.mDepth+1));
			if (interior->GetFront())
				tStack.push(BspTraversalData(interior->GetFront(), interior, frontPolys, data.mDepth+1));
		}
	}
}

Plane3 *BspTree::SelectPlane(Mesh *viewcell)
{
	return &viewcell->GetFacePlane(0);
}

BspNode *BspTree::SubdivideNode(BspLeaf *leaf, BspInterior *parent, 
								Mesh *viewCell, const int depth, 
								Mesh &frontPolys, Mesh &backPolys)
{
static int sMaxDepth = 5; // HACK
	// terminate traversal if no more faces in mesh
	if ((viewCell->mFaces.size() == 0) && depth > sMaxDepth)
		return leaf;
 
	// add the new nodes to the tree + select subdivision plane
	Plane3 *plane = SelectPlane(viewCell);
	BspInterior *node = new BspInterior(*plane); // ERROR!!
  

 	FaceContainer::const_iterator fi;

	for ( fi = viewCell->mFaces.begin(); fi != viewCell->mFaces.end(); ++ fi)
	{
	}
  
	BspLeaf *back = new BspLeaf(mRootCell);
	BspLeaf *front = new BspLeaf(mRootCell);

	// replace a link from node's parent
	if (parent)
	{
		parent->ReplaceChildLink(leaf, node);
	}

	// and setup child links
	node->SetupChildLinks(back, front);
  
	delete leaf;
	return node;
}

//} // GtpVisibilityPreprocessor